﻿# print lunisolar calendar from the current month on
#
# Rules:
# calendar is location specific
# days start and end at local sunset
# months start on a day when new moon is first visible at the location
# year starts at vernal equinox
# there are 3 spring months, 3 summer months, 3 autumn months and 3 - 4 winter months, so that spring equinox always happens in the first spring month

from __future__ import print_function

import ephem, sys, csv, codecs

from reportlab.lib.pagesizes import landscape, A4, cm
from reportlab.platypus import SimpleDocTemplate, Paragraph, Table, TableStyle, PageBreak, Spacer
from reportlab.lib import colors
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

try:
    from BytesIO import BytesIO
except ImportError:
    from io import BytesIO
    
    
class LuniSolarCalendar:

    def __init__(self, location):
        self.daysofweek = ['Po','Út','St','Čt','Pá','So','Ne']
        self.month_names = ['březen','duben','květen',
                            'červen','červenec','srpen',
                            'září','říjen','listopad',
                            'prosinec','leden','únor','hruden']
        
        self.s = ephem.Sun()
        self.m = ephem.Moon()
        
        self.loc = location

    def generate(self, months, output_format="csv"):
        self.output_format = output_format
            
        # get start of current day
        day = self.loc.previous_setting(self.s)

        # get previous vernal equinox
        ve = ephem.previous_vernal_equinox(day)

        # get previous new moon
        nm = ephem.previous_new_moon(ve)

        # adjust day for moon visibility
        day_b = self.adjust_new_month(nm)

        # if start of day is over equinox, start one new moon earlier
        if day_b > ve:
            day_b = adjust_new_month(ephem.previous_new_moon(nm))

        # get next vernal equinox from today to reset month numbers
        nve = ephem.next_vernal_equinox(day)

        month_number = 0
        
        month_count = 0
        
        self.print_start_cal()

        # loop
        while month_count < months:

            # get date of next new moon, starting tomorrow
            tomorrow = self.loc.next_setting(self.s, start = day_b)
            day_e = self.adjust_new_month(ephem.next_new_moon(tomorrow))
            
            # check if vernal equinox falls within the month
            if day_b <= nve and nve < day_e:
                # reset month number
                month_number = 0
                nve = ephem.next_vernal_equinox(day_e)
                
            # if end of this month is later than today, print it and count it (do not print historical months)
            if day_e >= day:
                # print out month
                self.process_month(day_b, day_e, month_number)
                # counter
                month_count = month_count + 1
        
            # increment month
            month_number = month_number + 1
            day_b = day_e

        self.print_end_cal()
        
        return self.output
        
    def adjust_new_month(self, nm):
        # if the crescent is observable during the day it occured, this day is first day of month, otherwise next day is
        
        today_b = self.loc.previous_setting(self.s, start = nm)
        today_e = self.loc.next_setting(self.s, start = nm)
        moonrise_l = self.loc.previous_rising(self.m, start = nm)
        moonset_l = self.loc.previous_setting(self.m, start = nm)
        moonrise_n = self.loc.next_rising(self.m, start = nm)
        
        # if moon is above horizon at new moon point, month starts today
        if moonrise_l > moonset_l:
            return today_b
        
        # otherwise if moon rises before sunset today, month starts today
        if moonrise_n < today_e:
            return today_b
        
        # otherwise month starts tomorrow
        return today_e
        
    def date_within(self, d, b, e):
        if b <= d and d < e:
            return True
        else:
            return False

    def process_month(self, start, end, month):

        self.print_start_month(month)
        
        d = start
        day_num = 1
        while (d + 0.01) < end:     # fix to rule out rounding errors when comparing on equality 
            # get start, transit and end of the date
            b = d
            e = self.loc.next_setting(self.s, start = d)
            t = self.loc.next_transit(self.s, start = d)
        
            # check astro phenomena: moon phases, sol/eq
            ephems = ""
            if self.date_within(ephem.next_new_moon(start), b, e):
                ephems += "new_moon "
            if self.date_within(ephem.next_first_quarter_moon(start), b, e):
                ephems += "first_quarter_moon "
            if self.date_within(ephem.next_full_moon(start), b, e):
                ephems += "full_moon "
            if self.date_within(ephem.next_last_quarter_moon(start), b, e):
                ephems += "last_quarter_moon "
            ns = ephem.next_solstice(start)
            if self.date_within(ns, b, e):
                if ns.datetime().month < 7:
                    ephems += "summer_"
                else:
                    ephems += "winter_"
                ephems += "solstice "
            if self.date_within(ephem.next_equinox(start), b, e):
                ephems += "equinox "
                
            # get local time
            lt = ephem.localtime(t)
                        
            # format civil date
            civil_date = str(lt.day) + "." + str(lt.month) + "."
            
            self.print_day(month, day_num, lt.weekday(), civil_date, ephems)
            
            # advance one day
            d = self.loc.next_setting(self.s, start = d)
            day_num = day_num + 1

        self.print_end_month()
        
    def print_start_cal(self):
        if self.output_format == "csv":
            self.print_start_cal_csv()
        if self.output_format == "pdf":
            self.print_start_cal_pdf()

    def print_end_cal(self):
        if self.output_format == "csv":
            self.print_end_cal_csv()
        if self.output_format == "pdf":
            self.print_end_cal_pdf()
        
    def print_start_month(self, month):
        if self.output_format == "csv":
            self.print_start_month_csv(month)
        if self.output_format == "pdf":
            self.print_start_month_pdf(month)
        
    def print_end_month(self):
        if self.output_format == "csv":
            self.print_end_month_csv()
        if self.output_format == "pdf":
            self.print_end_month_pdf()
        
    def print_day(self, month, day, weekday, civil_date, ephem):
        if self.output_format == "csv":
            self.print_day_csv(month, day, weekday, civil_date, ephem)
        if self.output_format == "pdf":
            self.print_day_pdf(month, day, weekday, civil_date, ephem)

    # CSV functions
    def print_start_cal_csv(self):
        # set up output
        self.output = StringIO()
        self.csv_writer = csv.DictWriter(self.output, fieldnames=['Month','Day','Weekday','Civil Date','Ephemeris'])
        self.csv_writer.writeheader()

    def print_end_cal_csv(self):
        content = self.output.getvalue()
        self.output.close()
        # convert to string
        self.output = content
        
    def print_start_month_csv(self, month):
        return
        
    def print_end_month_csv(self):
        return
        
    def print_day_csv(self, month, day, weekday, civil_date, ephem):
        # format month
        month_name = self.month_names[month]                        
        # format weekday
        weekday_name = self.daysofweek[weekday]

        self.csv_writer.writerow({'Month': month_name,'Day': day,'Weekday': weekday_name,'Civil Date': civil_date,'Ephemeris': ephem})
        return

    # PDF functions
    def print_start_cal_pdf(self):
        # set up output
        self.output = BytesIO()
        self.pdf_doc = SimpleDocTemplate(self.output, pagesize=landscape(A4), leftMargin=1*cm, rightMargin=1*cm, topMargin=1*cm, bottomMargin=1*cm)
        pdfmetrics.registerFont(TTFont('Inconsolata', 'Inconsolata.ttf'))
        pdfmetrics.registerFont(TTFont('astro', 'astro.ttf'))
        self.pdf_styles = getSampleStyleSheet()
        self.pdf_elements = []
        self.pdf_first_page = True


    def print_end_cal_pdf(self):
        self.pdf_doc.build(self.pdf_elements)
        content = self.output.getvalue()
        self.output.close()
        # convert to string
        self.output = content
        
    def print_start_month_pdf(self, month):
        if self.pdf_first_page == True:
            self.pdf_first_page = False
        else:
            self.pdf_elements.append(PageBreak())
        # month name
        month_name = self.month_names[month]
        style = self.pdf_styles["Title"]
        style.fontSize=40
        style.fontName ='Inconsolata'
        self.pdf_elements.append(Paragraph(month_name, style))
        # days of week
        self.pdf_elements.append(Spacer(1,1.2*cm))        
        data = []
        style.fontSize=20
        for d in self.daysofweek:
            data.append(Paragraph(d, style))
        table = Table([data], 7*[4*cm])
        self.pdf_elements.append(table)
        self.pdf_elements.append(Spacer(1,0.3*cm))        
        self.pdf_new_month = True
        self.pdf_data_table = []
        self.pdf_current_week = []

        
    def print_end_month_pdf(self):
        # finish last week and append to data
        for i in range(7 - len(self.pdf_current_week)):
            #empty cell
            self.pdf_current_week.append(self.pdf_cell_day())
        self.pdf_data_table.append(self.pdf_current_week)
        # calculate row height
        rh = 15.0 / len(self.pdf_data_table)
        # output calendar table
        table = Table(self.pdf_data_table,7*[4*cm], len(self.pdf_data_table)*[rh*cm])
        table.setStyle(TableStyle([
                       ('INNERGRID', (0,0), (-1,-1), 0.25, colors.black),
                       ('BOX', (0,0), (-1,-1), 0.25, colors.black),
                       ('LEFTPADDING', (0,0), (-1,-1), 0),
                       ('RIGHTPADDING', (0,0), (-1,-1), 0),
                       ('BOTTOMPADDING', (0,0), (-1,-1), 0),
                       ('TOPPADDING', (0,0), (-1,-1), 0),
                       ('VALIGN', (0,0), (-1,-1), 'TOP'),
                       ]))
        self.pdf_elements.append(table)
        
    def print_day_pdf(self, month, day, weekday, civil_date, ephem):
        #if new month, fill in days
        if self.pdf_new_month == True:
            self.pdf_new_month = False
            for i in range(weekday):
                #empty cell
                self.pdf_current_week.append(self.pdf_cell_day())
        if len(self.pdf_current_week) == 7:
            self.pdf_data_table.append(self.pdf_current_week)
            self.pdf_current_week = []
        self.pdf_current_week.append(self.pdf_cell_day(day, civil_date, ephem))
                                
        # format weekday
        weekday_name = self.daysofweek[weekday]
    
    def pdf_cell_day(self,    day = '', civil_date = '', ephem = ''):
        ephem_string = ''
        if 'new_moon' in ephem:
            ephem_string += 'B '
        if 'first_quarter_moon' in ephem:
            ephem_string += 'G '
        if 'full_moon' in ephem:
            ephem_string += 'A '
        if 'last_quarter_moon' in ephem:
            ephem_string += 'H '
        if 'equinox' in ephem:
            ephem_string += 'KJ '
        if 'summer_solstice' in ephem:
            ephem_string += 'J '
        if 'winter_solstice' in ephem:
            ephem_string += 'K '
        row = [ephem_string]
        row.append(day)
        data = [row]
        row = ['']
        row.append(civil_date)
        data.append(row)
        table = Table(data, 2*[2*cm])
        table.setStyle(TableStyle([
                       ('ALIGN', (-1,0), (-1,-1), 'RIGHT'),
                       ('VALIGN', (0,0), (-1,0), 'TOP'),
                       ('VALIGN', (0,-1), (-1,-1), 'BOTTOM'),
                       ('SIZE', (-1,0), (-1,0), 1*cm),
                       ('LEADING', (-1,0), (-1,0), 1*cm),                       
                       ('SIZE', (-1,-1), (-1,-1), 0.2*cm),
                       ('TEXTCOLOR', (-1,-1), (-1,-1), colors.grey),
                       ('BOTTOMPADDING', (-1,-1), (-1,-1), 0.1),
                       ('FONT', (0,0), (0,0), 'astro'),
                       ('TOPPADDING', (0,0), (0,0), 8),
                       ]))
        return table
        


# MAIN

def main(argv=None):
    if len(sys.argv) == 1:
        print()
        print("Usage:", sys.argv[0], "<city> <month count> {csv|pdf} <filename>")
        print()
        print("Example:", sys.argv[0], "Geneva 12 pdf calendar.pdf")
        print()
        return
    else:
        city = sys.argv[1]
        months = int(sys.argv[2])
        format = sys.argv[3]
        fname = sys.argv[4]

    # location
    loc = ephem.city(city)

    # calendar object
    cal = LuniSolarCalendar(loc)
    
    data = cal.generate(months, format)
    
    with open(fname, 'wb') as file:
        # convert string buffers to bytes
        try:
            data = data.encode()
            file.write(codecs.BOM_UTF8)
        except:
            pass
        file.write(data)
        file.close()
    

    
if __name__ == "__main__":
    sys.exit(main())